#! /bin/sh
#
# srun.sh can be called like apachectl
#
# srun.sh         -- execs the servlet runner in the foreground
# srun.sh start   -- starts the servlet runner in the background
# srun.sh stop    -- stops the running servlet runner
# srun.sh restart -- restarts the running servlet runner
#
# srun.sh will return a status code if the wrapper detects an error, but
# some errors, like bind exceptions or Java errors, are not detected.
#
# Customized arguments, e.g. -resin_home or -java_home or -pid.
#
# -pid <pidfile>         -- use a non-default pid file
#                           (useful for multiple servers)
# -java_home <java_home> -- use a non-default Java home
# -stdout <filename>     -- stdout message log
# -stderr <filename>     -- stderr message log
# -native                -- force native threads
# -green                 -- force green threads
# -verbose               -- prints Java arguments before starting.
# -no-auto-restart       -- disable automatic server restart
#                        -- (this only appled to start and restart)
#
# Extra arguments.  If you're passing arguments to the JVM, you'll
# need to use -Jxxx.  For example, -J-ms48m.
#
# If you're using a primary/backup on the same machine, you may want to
# create two srun scripts with "-pid srun1.pid" and "-pid srun2.pid" in
# the args.
#
args=
#
# class to start
#
class=com.caucho.server.http.RunnerServer
#
# name of the server
#
name=srun
#
# location of perl executable
#
perl=perl

#
# trace script and simlinks to find the wrapper
#
script=`/bin/ls -l $0 | awk '{ print $NF; }'`

while test -h "$script"
do
  script=`/bin/ls -l $script | awk '{ print $NF; }'`
done

bin=`dirname $script`

exec $perl $bin/wrapper.pl -chdir -name "$name" -class "$class" $args $*


